import csv
import os
import shutil
import socket
import urllib
from datetime import datetime, timedelta
import numpy as np

import matplotlib as mpl
mpl.use('QT5Agg')
from matplotlib import pyplot as plt
import matplotlib.patches as mplpatches


class PerformanceTHREDDS(object):

    def __init__(self):
        self._thredds_time = list()
        self._copy_time = list()
        self.total_attempts = 100

        self.scratch_folder = '/scratch/Earth/jvegas/THREDDS_test/'
        self.results_folder = 'esnas'
        self.url_thredds = 'http://earth.bsc.es/thredds/fileServer/exp/ecmwf/system4_m1/6hourly/tas/tas_19810101.nc'
        self.data_path = '/esnas/exp/ecmwf/system4_m1/6hourly/tas/tas_19810101.nc'

        self.file_size = None
        self.node = None

        self.marker_size = 6

    def test_performance(self):
        print('Testing performance')
        print('THREDDS test url: {0}'.format(self.url_thredds))
        print('Copy file path:   {0}'.format(self.data_path))

        self.node = socket.gethostname()
        self.file_size = os.path.getsize(self.data_path)
        self.tests_start = datetime.now()

        if not os.path.isdir(self.scratch_folder):
            os.makedirs(self.scratch_folder)

        for x in range(self.total_attempts):
            print ('Test {0} of {1}'.format(x+1, self.total_attempts))
            self.test_thredds()
            self.test_copy()

        self.tests_end = datetime.now()

        print('Tests finished!')

        self._prepare_data()

    def _prepare_data(self):
        self._copy_time = np.array(self._copy_time)
        self._copy_average = np.average(self._copy_time)
        self._copy_std = np.std(self._copy_time)

        self._thredds_time = np.array(self._thredds_time)
        self._thredds_average = np.average(self._thredds_time)
        self._thredds_std = np.std(self._thredds_time)

        self._attempts = np.array(range(self.total_attempts))

    def test_thredds(self):
        start = datetime.now()
        urllib.urlretrieve(self.url_thredds, os.path.join(self.scratch_folder, 'thredds.nc'))
        ellapsed = datetime.now() - start
        self._thredds_time.append(ellapsed.total_seconds())
        print 'Tested THREDDS: {0}'.format(ellapsed)

    def test_copy(self):
        start = datetime.now()
        shutil.copyfile(self.data_path, os.path.join(self.scratch_folder, 'copy.nc'))
        ellapsed = datetime.now() - start
        self._copy_time.append(ellapsed.total_seconds())
        print 'Tested copy: {0}'.format(ellapsed)

    def save_results(self):
        csv_path = self._get_result_filepath('csv')
        print ('Saving result at {0}'.format(csv_path))
        with open(csv_path, 'wb') as csvfile:
            spamwriter = csv.writer(csvfile)
            spamwriter.writerow(['File', 'Filesize', 'Node'])
            spamwriter.writerow([self.data_path, self.file_size, self.node])
            spamwriter.writerow(['Attempt', 'THREDDS', 'Copy'])
            for x in range(self.total_attempts):
                spamwriter.writerow([x, self._thredds_time[x], self._copy_time[x]])
        print('Saving done!')

    def load_results(self, filepath):
        if not os.path.isfile(filepath):
            filepath = os.path.join(self.results_folder, filepath)
        with open(filepath, 'r') as csvfile:
            reader = csv.reader(csvfile)
            row = reader.next()
            if row[0] == 'File':
                self.data_path, self.file_size, self.node = reader.next()
                self.file_size = float(self.file_size)
                reader.next()
            for row in reader:
                self._thredds_time.append(float(row[1]))
                self._copy_time.append(float(row[2]))
            self.total_attempts = len(self._thredds_time)

        self.tests_start = datetime.strptime(os.path.basename(filepath)[-20:-4], '%Y-%m-%d-%H:%M')
        test_time = timedelta()
        for seconds in self._thredds_time:
            test_time += timedelta(seconds=seconds)
        for seconds in self._copy_time:
            test_time += timedelta(seconds=seconds)
        self.tests_end = self.tests_start + test_time

        self._prepare_data()

    def plot(self, save):
        print('Plotting results...')
        fontsize = 20
        fig, ax1 = plt.subplots(figsize=(16, 9))

        if self.file_size is None:
            self.file_size = os.path.getsize(self.data_path)

        fig.suptitle('Comparison between THREDDS and direct copy performance on {0} '
                     '({1:%b %d %Y} {1:%H:%M} - {2:%H:%M}) \n'
                     'File used: {3} ({4:0.2f} GB)'.format(self.node, self.tests_start, self.tests_end, self.data_path,
                                                      self.file_size / (1024 * 1024 * 1024.0)),
                     fontsize=fontsize)

        ax1.plot(self._attempts, self._copy_time, 'bo', markersize=self.marker_size)
        plt.axhspan(self._copy_average - self._copy_std, self._copy_average + self._copy_std, facecolor='b', alpha=0.3)
        plt.axhline(y=self._copy_average, color='b')
        ax1.plot(self._attempts, self._thredds_time, 'ro', markersize=self.marker_size)
        plt.axhspan(self._thredds_average - self._thredds_std, self._thredds_average + self._thredds_std, facecolor='r', alpha=0.3)
        plt.axhline(y=self._thredds_average, color='r')

        def time_ticks(x, pos):
            d = timedelta(seconds=x)
            return str(d)

        ax1.set_ylabel('Time', fontsize=fontsize)
        ax1.set_xlabel('Attempt', fontsize=fontsize)
        ax1.set_xlim([-0.5, self.total_attempts - 0.5])

        max_time = 60
        while np.amax(self._copy_time) > max_time or np.amax(self._thredds_time) > max_time:
            max_time += 10
        ax1.set_ylim([0, max_time])
        formatter = mpl.ticker.FuncFormatter(time_ticks)
        ax1.yaxis.set_ticks(np.arange(0, max_time, 30))
        ax1.yaxis.set_ticks(np.arange(0, max_time, 10), minor=True)
        ax1.yaxis.set_major_formatter(formatter)
        ax1.yaxis.grid(which='major', color='black', linestyle='-')
        ax1.yaxis.grid(which='minor', color='grey', linestyle='--')

        plt.tick_params(axis='x', which='both', bottom='off', top='off', labelbottom='off')

        copy_patch = mplpatches.Patch(color='b', label='Copy')
        thredds_patch = mplpatches.Patch(color='r', label='THREDDS')
        ax1.legend(handles=[copy_patch, thredds_patch], loc='upper center', bbox_to_anchor=(0.5, 1),
                   fancybox=True, ncol=2, frameon=True, fontsize=14)
        if save:
            print('Saving image...')
            fig.savefig(self._get_result_filepath('png'))
        else:
            print('Showing image...')
            plt.show()
        plt.close()
        print('Plotting finished!')

    def _get_result_filepath(self, extension):
        if not os.path.isdir(self.results_folder):
            os.makedirs(self.results_folder)
        return os.path.join(self.results_folder, 'cp_{2}_{0:%Y-%m-%d-%H:%M}.{1}'.format(self.tests_start, extension,
                                                                                      self.node))
